program S_Delphi;

{ author:    Michael Hieke, 1997, 1998
  e-mail:    MHieke42@aol.com
  purpose:   Starts DELPHI.EXE and - if running under Windows95 - hooks into
             creation of Delphi's forms to apply the WS_EX_TOOLWINDOW style.
             This way they don't appear in the task bar, only the main form does.
             Keeps track of LoadLibrary() calls and frees these DLL's after crash.
}

{$I settings.inc}
{$R S_DELPHI.RES}

{$IFDEF WIN32}
  szWarning = 'Don''t try this at home!';
{$ELSE}

uses
  WinTypes,
  WinProcs,
  Messages,
  Classes,
  SysUtils,
  PrcHook in 'PrcHook.PAS', { intercept API calls }
  ToolHelp,                 { system notifications }
  U_llvl in 'U_LLVL.PAS';   { prog's low level stuff }

const
  AppName          = 'Start_Delphi1';

var
  wndMain: HWND;
{$IFDEF DEBUG}
  szBuf: array[0..300] of char;
{$ENDIF}

{------------------------------------------------------------------------------}

const
  aDelphiModules: TStringList = nil;
  aLoadLibModules: TStringList = nil;
  ghDelphiModule: THandle = 0;

procedure HandleLastModuleRemoved; forward;

procedure HandleEvent(pEvt: PNFYEvent);
var sModule: string;
    nIdx: integer;
begin
    { if it's some new module, get name and store it, because when the event
      NFY_DELMODULE gets here you can't call GetModuleFileName() because the
      module handle has been invalidated (PostMessage() from handler) }
  case pEvt^.wID of
    NFY_STARTTASK:
      begin
        Byte(sModule[0]):= GetModuleFileName(pEvt^.hMod, @sModule[1], 255);
{$IFDEF DEBUG}
        StrFmt(szBuf, 'start task "%s" (module %4xh)'#10,
                      [sModule, pEvt^.hMod]);
        OutputDebugString(szBuf);
{$ENDIF}
          { store this module handle in special variable, because it's the
            last module to be removed }
        ghDelphiModule:= pEvt^.hMod;
          { add to list of loaded modules }
        aDelphiModules.AddObject(sModule, pointer(pEvt^.hMod));
      end;
{$IFDEF DEBUG}
    NFY_EXITTASK:
      begin
          { seek in list of loaded modules }
        nIdx:= aDelphiModules.IndexOfObject(pointer(pEvt^.hMod));
        if (nIdx >= 0) then begin
          StrFmt(szBuf, 'kill task "%s" (module %4xh)'#10,
                        [aDelphiModules[nIdx], pEvt^.hMod]);
        end else begin
          StrFmt(szBuf, 'error: kill task with module %4xh which was not loaded'#10,
                        [pEvt^.hMod]);
        end;
        OutputDebugString(szBuf);
      end;
{$ENDIF}
    NFY_STARTDLL:
      begin
        Byte(sModule[0]):= GetModuleFileName(pEvt^.hMod, @sModule[1], 255);
{$IFDEF DEBUG}
        StrFmt(szBuf, 'load dll "%s" (module %4xh)'#10,
                      [sModule, pEvt^.hMod]);
        OutputDebugString(szBuf);
{$ENDIF}
          { add to list of loaded modules }
        aDelphiModules.AddObject(sModule, pointer(pEvt^.hMod));
      end;
    NFY_DELMODULE:
      begin
          { seek in list of loaded modules and remove }
        nIdx:= aDelphiModules.IndexOfObject(pointer(pEvt^.hMod));
        if (nIdx >= 0) then begin
{$IFDEF DEBUG}
          StrFmt(szBuf, 'unload module "%s" (module %4xh)'#10,
                        [aDelphiModules[nIdx], pEvt^.hMod]);
          OutputDebugString(szBuf);
{$ENDIF}
          aDelphiModules.Delete(nIdx);
            { if this was the last module to be removed }
          if (pEvt^.hMod = ghDelphiModule) then HandleLastModuleRemoved;
        end else begin
{$IFDEF DEBUG}
          StrFmt(szBuf, 'error: unload module %4xh which was not loaded'#10,
                        [pEvt^.hMod]);
          OutputDebugString(szBuf);
{$ENDIF}
        end;
      end;
{$IFDEF DEBUG}
    NFY_MH_DEBUGOUT:
      if (pEvt^.atomMsg <> 0) then begin
        if (GlobalGetAtomName(pEvt^.atomMsg, szBuf, SizeOf(szBuf) - 1) > 0) then
          OutputDebugString(szBuf);
        GlobalDeleteAtom(pEvt^.atomMsg);
      end;
{$ENDIF}
    NFY_MH_LOADLIBRARY:
      if (pEvt^.atomMsg <> 0) then begin
        Byte(sModule[0]):= GlobalGetAtomName(pEvt^.atomMsg, @sModule[1], 255);
        aLoadLibModules.AddObject(sModule, pointer(pEvt^.hMod));
        GlobalDeleteAtom(pEvt^.atomMsg);
{$IFDEF DEBUG}
        StrFmt(szBuf, 'LoadLibrary("%s"): %4xh'#10, [sModule, pEvt^.hMod]);
        OutputDebugString(szBuf);
{$ENDIF}
      end;
    NFY_MH_FREELIBRARY:
      begin
        nIdx:= aLoadLibModules.IndexOfObject(pointer(pEvt^.hMod));
        if (nIdx > -1) then begin
{$IFDEF DEBUG}
          StrFmt(szBuf, 'FreeLibrary("%s")'#10, [aLoadLibModules[nIdx]]);
          OutputDebugString(szBuf);
{$ENDIF}
          aLoadLibModules.Delete(nIdx);
        end;
      end;
  end;
  Dispose(pEvt);
end;

procedure HandleEvents;
var pEvt: PNFYEvent;
begin
  while TRUE do begin
    pEvt:= GetNextPendingEvent;
    if (pEvt = nil) then break;
    HandleEvent(pEvt);
  end;
end;

procedure HandleLastModuleRemoved;
begin
  if (aLoadLibModules.Count > 0) then begin
{$IFDEF DEBUG}
    StrFmt(szBuf, 'error: missing FreeLibrary() for %d module(s)'#10,
                  [aLoadLibModules.Count]);
    OutputDebugString(szBuf);
{$ENDIF}
    PostMessage(wndMain, WM_REMOVEDLLS, 0, 0);
  end else
    if IsWindow(wndMain) then PostMessage(wndMain, WM_CLOSE, 0, 0);
end;

procedure TryRemoveDLLs;
var i: integer;
    hMod: THandle;
begin
  for i:= Pred(aLoadLibModules.Count) downto 0 do begin
{$IFDEF DEBUG}
    StrFmt(szBuf, 'calling FreeLibrary() for "%s"'#10, [aLoadLibModules[i]]);
    OutputDebugString(szBuf);
{$ENDIF}
    hMod:= THandle(aLoadLibModules.Objects[i]);
    FreeLibrary(hMod);
    aLoadLibModules.Delete(i);
  end;
  if IsWindow(wndMain) then PostMessage(wndMain, WM_CLOSE, 0, 0);
end;

{------------------------------------------------------------------------------}

function WindowProc(wnd: HWND; Msg, wParam: word; lParam: longint): longint; export;
begin
  Result:= 0;
  case Msg of
{$IFDEF DEBUG}
    WM_LBUTTONDOWN:
      if (GetDelphiTask <> 0) then TerminateApp(GetDelphiTask, NO_UAE_BOX);
{$ENDIF}
    WM_DESTROY: PostQuitMessage(0);
    WM_MODULELISTCHANGED: HandleEvents;
    WM_REMOVEDLLS: TryRemoveDLLs;
    else Result:= DefWindowProc(wnd, Msg, wParam, lParam);
  end;
end;

{------------------------------------------------------------------------------}

procedure RegisterWndClass;
const
  wndClass: TWndClass = (
              Style: 0;
              lpfnWndProc: @WindowProc;
              cbClsExtra: 0;
              cbWndExtra: 0;
              hInstance: 0;
              hIcon: 0;
              hCursor: 0;
              hbrBackground: 0;
              lpszMenuName: nil;
              lpszClassName: AppName
            );
begin
  wndClass.hInstance:= HInstance;
{$IFDEF DEBUG}
  wndClass.hbrBackground:= GetStockObject(WHITE_BRUSH);
  wndClass.hCursor:= LoadCursor(0, IDC_ARROW);
{$ENDIF}
  if not WinProcs.RegisterClass(wndClass) then Halt(255);
end;

procedure WinMain;
var wnd: HWND;
    Msg: TMSG;
begin
  if (HPrevInst <> 0) then exit;
    { just start Delphi and exit if not Windows95 }
  if (Word(GetVersion) < $5F03) then begin
    WinExec('DELPHI.EXE', SW_SHOW);
    Halt;
  end;

  aDelphiModules:= TStringList.Create;
  aLoadLibModules:= TStringList.Create;

    { register window class and create hidden main window }
  RegisterWndClass;
  wndMain:= CreateWindow(AppName, '', WS_OVERLAPPEDWINDOW,
{$IFDEF DEBUG}
                         GetSystemMetrics(SM_CXSCREEN) - 120,
                         GetSystemMetrics(SM_CYSCREEN) - 80, 120, 80,
{$ELSE}
                         CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
{$ENDIF}
                         0, 0, HInstance, nil);

{$IFDEF DEBUG}
  ShowWindow(wndMain, SW_SHOW);
  UpdateWindow(wndMain);
{$ENDIF}

    { get notified of special app events }
  NotificationStart(wndMain);
    { hook into CreateWindowEx() API }
  APIInterceptionStart;

    { start Delphi }
  if (WinExec('DELPHI.EXE', SW_SHOW) < HINSTANCE_ERROR) then begin
    APIInterceptionStop;
    NotificationStop;
    DestroyWindow(wndMain);
    aDelphiModules.Free;
    aLoadLibModules.Free;
{$IFDEF DEBUG}
    OutputDebugString('Unable to execute DELPHI.EXE'#10);
{$ENDIF}
    Halt;
  end;

  while GetMessage(Msg, 0, 0, 0) do begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;

    { clean up things }
  APIInterceptionStop;
  HandleEvents;    { restoring the window procedure may generate an event }
  NotificationStop;
  aDelphiModules.Free;
  aLoadLibModules.Free;

  Halt(Msg.wParam);
end;

begin
  WinMain;
{$ENDIF} {WIN32}
end.
